"""
Database-guided multi-robot motion planning
"""
import random
import pygame
import os
from shapely.geometry import Polygon, Point



from guided_mrmp.utils import Env
from guided_mrmp.utils.helpers import *

from guided_mrmp.conflict_resolvers import TrajOptDBResolver

from guided_mrmp.utils import Roomba


T = 1  # Prediction Horizon [s]
DT = 0.2  # discretization step [s]

class GuidedMRMP:
    def __init__(self, env, robots, dynamics_models):
        """
        inputs:
            - robots (list): list of Robot class objects
            - env (Env): the environment
        """
        self.robots = robots
        self.dynamics_models = dynamics_models
        self.env = env
        self.current_guides = []*len(robots)
        self.current_trajs = []*len(robots)


    def find_all_conflicts(self, dt):
        """
        Loop over all the robots, checking for both node conflicts and edge
        conflicts. 
        """
        conflicts = []

        for idx, r1 in enumerate(self.robots):
            control = r1.next_control
            next_state = self.dynamics_models[idx].next_state(r1.current_position, control, dt)
            circ1 = Point(next_state[0],next_state[1])
            circ1 = circ1.buffer(r1.radius)
            for r2 in self.robots:
                if r1.label == r2.label:
                    continue

                control = r2.next_control
                next_state = self.dynamics_models[idx].next_state(r2.current_position, control, dt)
                circ2 = Point(next_state[0],next_state[1])
                circ2 = circ2.buffer(r2.radius)
                
                # check if the robots overlap
                if circ1.intersects(circ2): 
                    if (r1,r2) not in conflicts and (r2,r1) not in conflicts:
                        print(f"conflict between {r1.label} and {r2.label}")
                        conflicts.append((r1,r2))

        return conflicts

    def all_robots_at_goal(self):
        for r in self.robots:
            if (np.sqrt((r.tracker.state[0] - r.goal[0]) ** 2 + (r.tracker.state[1] - r.goal[1]) ** 2) > 0.1):
                return False
        return True

    def advance(self, screen, state, time, dt=0.1):
        """
        Advance the simulation by one timestep.

        inputs:
            - state (list): list of tuples, where each tuple is the state of a robot
            - time (float): current time
            - t (float): time step
        """

        # get the next control for each robot
        for r in self.robots:
            r.next_control = r.tracker.get_next_control(r.current_position)

        # find all the conflicts at the next timestep
        conflicts = self.find_all_conflicts(dt)

        # resolve the conflicts using the database
        resolver = TrajOptDBResolver()
        resolver.resolve_conflict(conflicts, screen)


        # update the state of each robot

        # for idx, r in enumerate(self.robots):
        #     control = r.next_control
        #     r.current_position = self.dynamics_models[idx].next_state(r.current_position, control, dt)
        #     # self.current_trajs[idx].append(r.state)
            
        #     r.x_history.append(r.state[0])
        #     r.y_history.append(r.state[1])
        #     r.h_history.append(r.state[2])

        # return the valid controls
        controls = [r.next_control for r in self.robots]
        return controls

    def add_vis_guide_paths(self, screen):
        """
        Add the visualization to the screen.
        """
        
        for r in self.robots:
            path = r.rrtpath
            for node in path:
                pygame.draw.circle(screen, r.color, (int(node[0]), int(node[1])), 2)
            for i in range(len(path)-1):
                pygame.draw.line(screen, r.color, r.rrtpath[i], r.rrtpath[i+1], 2)

    def add_vis_grid(self, screen, grid_size, top_left, cell_size):
        """
        Add the visualization to the screen.
        """
         # draw the grid
        line_len = grid_size*cell_size
        for i in range(grid_size+1):
            xs = [top_left[0] + i*cell_size,top_left[0] + i*cell_size]
            ys = [top_left[1], top_left[1]-line_len]
            pygame.draw.line(screen, (0,0,0), (xs[0],ys[0]), (xs[1],ys[1]), 2)

            xs = [top_left[0], top_left[0]+line_len]
            ys = [top_left[1] - i*cell_size,top_left[1] - i*cell_size]
            
            pygame.draw.line(screen, (0,0,0), (xs[0],ys[0]), (xs[1],ys[1]), 2)

if __name__ == "__main__":


    # set_python_seed(42)

    # create the environment
    env = Env()

    # starts, goals = create_random_starts_and_goals(env, 4)
    starts = [[0,0.1,0],[10,5.1,0]]
    goals = [[10,3,0],[5,1,0]]

    alg = GuidedMRMP(Env)
    alg.run(starts, goals)